package edu.odu.cs.zeil.AlgAE.Snapshot;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.NoSuchElementException;

import edu.odu.cs.zeil.AlgAE.Client.SourceViewer.SourceLocation;
import edu.odu.cs.zeil.AlgAE.Snapshot.Rendering.Renderer;

/**
 * A snapshot is a model of the current data state in terms of a graph of connected
 * Entities.  The graph is "rooted" at an ActivationStack, which provides access
 * to
 *   1) Parameters for each active function call
 *   2) A list of entities that are local to the topmost activation (most recent call)
 *   3) A list of entities that are global to all activations
 * The complete graph is formed from the transitive closure over the component and connection
 * relations starting from that activation stack.
 * 
 * @author zeil
 *
 */
public class Snapshot  implements Iterable<Entity> {
	
	/**
	 *  Provides access to all objects in this snapshot, indexed by their identifiers.
	 *  
	 */
	private HashMap<Identifier, LinkedList<Entity> > entities;

	/**
	 * The activation stack, which forms the root from which all other entities in the 
	 * model can be located. 
	 */
	protected ActivationStack astack;
	
	
	/**
	 * An informational string to appear in a status line when this snapshot is being displayed 
	 */
	protected String descriptor;
	
	/**
	 * The corresponding source code location to be shown while this snapshot is being displayed.
	 */
	protected SourceLocation sourceLoc;
	
	
	private EntityIdentifier stackEntity;
	
	
	
	/**
	 * Create a new snapshot.
	 * @param description
	 * @param loc
	 */
	public Snapshot(String description, SourceLocation loc, ActivationStack callStack) {
		entities = new HashMap<Identifier, LinkedList<Entity>>();
		descriptor = description;
		sourceLoc = (loc != null) 
			? loc
			: new SourceLocation(callStack.getClass(), "", 0);
		astack = callStack;
		
		formClosure();
		normalize();
	}
	


	/**
	 * Guards against infinite recursion in the componentOf relation
	 */
	private static final int DepthLimit = 25;
	
	
	private class InternalComponent {
		public EntityIdentifier container;
		public Component component;
		
		public InternalComponent (EntityIdentifier container, Component component)
		{
			this.container = container;
			this.component = component;
		}
		
		public String toString() {
			return "IC[" + component + "]@" + container;
		}
	}
	
	
	
	/**
	 * Adds to the snapshot all objects that can be reached in one or more
	 * steps along the component and connection from the activation stack. This
	 * will include all global variables, parameters and locals of the current call,
	 * and non-ref parameters of the older calls.
	 *  
	 * @param ob
	 * @param tobeProcessed
	 */
	private void formClosure()
	{
		// Initialize the queue with the activation stack and with the global and
		// local objects
		LinkedList<InternalComponent> queue = new LinkedList<InternalComponent>();
		stackEntity = new EntityIdentifier(Identifier.getIdentifierFor(astack, astack.context()));
		queue.add (new InternalComponent(null, new Component(astack)));
		
		List<Component> globals = astack.getGlobalComponents();
		for (Component gC: globals) {
			queue.add(new InternalComponent(null, gC));
		}

		while (!queue.isEmpty()) {
			// For each component in the queue, use the renderer for that object
			// to create an entity and add its components and connections to the queue
			// for future processing.
			InternalComponent c = queue.pop();
			Identifier oid = Identifier.getIdentifierFor(c.component.getComponentObject(), astack.context());
			LinkedList<Entity> aliasesForEntity = entities.get(oid);
			if (aliasesForEntity == null) {
				aliasesForEntity = new LinkedList<Entity>();
				entities.put(oid, aliasesForEntity);
			}
			boolean found = false;
			EntityIdentifier newEntityID = new EntityIdentifier(oid, c.container, c.component.getLabel());
			//System.err.println ("Snapshot closure adds " + newEntityID);

			for (Entity e: aliasesForEntity) {
				if (newEntityID.equals(e.getEntityIdentifier())) {
					found = true;
					break;
				}
			}
			if (!found) {
				Entity entity = renderObject (newEntityID, c, queue);
				//System.err.println ("Closure: new entity " + entity);
				aliasesForEntity.add(entity);
			}
		}
	}

	// Create an entity describing the rendering of this component;
	private Entity renderObject(EntityIdentifier eid, InternalComponent c, LinkedList<InternalComponent> queue) {
		Identifier oid = eid.getObjectIdentifier();
		Entity entity = null;
		if (c.container == null) {
			String label = c.component.getLabel();
			entity = new Entity(oid, label);
		} else {
			entity = new Entity(oid, c.container, c.component.getLabel());
		}
		Object obj = c.component.getComponentObject();
		Renderer<Object> render = astack.getRenderer(obj);
		entity.setColor(render.getColor(obj));
		entity.setMaxComponentsPerRow(render.getMaxComponentsPerRow(obj));
		entity.setValue(render.getValue(obj));
		List<Component> components = render.getComponents(obj);
		if (components != null && eid.depth() < DepthLimit) {
			int componentCount = 0;
			for (Component comp: components) {
				Object cobj = comp.getComponentObject();
				String clabel = comp.getLabel();
				if (clabel == null || clabel.length() == 0)
					clabel = "\t" + componentCount;
				++componentCount;
				if (cobj != null) {
					EntityIdentifier c_eid = new EntityIdentifier(Identifier.getIdentifierFor(cobj, astack.context()), eid, clabel);
					entity.getComponents().add(c_eid);
					InternalComponent intComp = new InternalComponent(eid, new Component(cobj, clabel));
					//System.err.println ("" + entity.getEntityIdentifier() + " has component " + c_eid);
					queue.add (intComp);
				}
			}
		}
		List<Connection> connections = render.getConnections(obj);
		if (connections != null) {
			for (Connection conn: connections) {
				Object destObj = conn.getDestination();
				Identifier destID = null;
				destID = Identifier.getIdentifierFor(destObj, astack.context());
				Connector connector = new Connector(conn.getID(), eid, new EntityIdentifier(destID), 
						conn.getMinAngle(), conn.getMaxAngle(), conn.getComponentIndex());
				if (conn.getColor() != null)
					connector.setColor(conn.getColor());
				connector.setLabel(conn.getLabel());
				connector.setValue(conn.getValue());
				connector.setElasticity(conn.getElasticity());
				connector.setPreferredLength(conn.getPreferredLength());
				entity.getConnections().add(connector);
				if (destObj != null) {
					InternalComponent intComp = new InternalComponent(null, new Component(destObj));
					//System.err.println ("" + entity.getEntityIdentifier() + " connects to " + destID);
					queue.add (intComp);
				}
			}
		}
		return entity;
	}


	
	/**
	 * Attempts to resolve duplications caused by objects that map onto several discrete entities.
	 * 
	 * 1) If two entities exist for the same object and one is not a component of a larger entity
	 *    and is also unlabeled, then that one is removed.
	 * 2) For each remaining object with multiple renderings, the most deeply nested one is considered as
	 *      the primary occurrence. If there is a tie for most deeply nested, the tie is broken arbitrarily.
	 * 3) All connectors incoming to an entity are re-routed to the primary occurrence of that object.
	 * 4) All connectors outgoing from a non-primary entity are dropped.
	 */
	private void normalize()
	{
		HashMap<Identifier, Entity> primaries = new HashMap<Identifier, Entity>();
		HashMap<EntityIdentifier, Entity> unique = new HashMap<EntityIdentifier, Entity>();
		for (Identifier oid: entities.keySet()) {
			LinkedList<Entity> aliases = entities.get(oid);
			int maxdepth = -1;
			Entity deepest = null;
			Iterator<Entity> it = aliases.iterator();
			while (it.hasNext()) {
				Entity entity = it.next();
				unique.put (entity.getEntityIdentifier(), entity);
				if (aliases.size() > 1) {
					if (entity.getContainer() == null && 
							(entity.getLabel() == null ||  entity.getLabel().length() == 0)) {
						it.remove();
					}		
				}
				int deep = entity.getEntityIdentifier().depth();
				if (deep > maxdepth) {
					maxdepth = deep;
					deepest = entity;
				}
			}
			primaries.put(oid, deepest);
		}
		
		HashSet<EntityIdentifier> keepThese = new HashSet<EntityIdentifier>();
		LinkedList<Entity> queue = new LinkedList<Entity>();
		for (Identifier oid: entities.keySet()) {
			LinkedList<Entity> aliases = entities.get(oid);
			for (Entity e: aliases) {
				EntityIdentifier ceid = e.getContainer(); 
				if (ceid == null || ceid.equals(EntityIdentifier.nullID())) {
					queue.add(e);
				}
			}
		}
		while (!queue.isEmpty()) {
			Entity e = queue.getFirst();
			queue.removeFirst();
			EntityIdentifier eid = e.getEntityIdentifier();
			keepThese.add(eid);
			for (EntityIdentifier ceid: e.getComponents()) {
				queue.add (unique.get(ceid));
			}
		}
		HashMap<Identifier, LinkedList<Entity>> trimmedEntities = new HashMap<Identifier, LinkedList<Entity>>();
		for (Identifier oid: entities.keySet()) {
			LinkedList<Entity> aliases = entities.get(oid);
			Iterator<Entity> it = aliases.iterator();
			while (it.hasNext()) {
				Entity entity = it.next();
				if (!keepThese.contains(entity.getEntityIdentifier())) {
					it.remove();		
				}
			}
			if (aliases.size() > 0) {
				trimmedEntities.put (oid, aliases);
			}
		}
		entities = trimmedEntities;
		
		for (Identifier oid: entities.keySet()) {
			LinkedList<Entity> aliases = entities.get(oid);
			Entity primary = primaries.get(oid);
			for (Entity e: aliases) {
				if (e == primary) {
					for (Connector conn: e.getConnections()) {
						EntityIdentifier destID = conn.getDestination();
						if (!destID.equals(EntityIdentifier.nullID())) {
							EntityIdentifier primaryDest = primaries.get(destID.getObjectIdentifier()).getEntityIdentifier();
							conn.setDestination (primaryDest);
						}
					}
				} else {
					e.getConnections().clear();
				}
			}
		}		
	}

		
	public String toString()
	{
		return "" + descriptor + ":" + sourceLoc.toString() + "["
		+ entities.size() + " entities]";
	}

	



	/**
	 * An informational string to appear in a status line when this snapshot is being displayed 
	 * 
	 * @param descriptor the descriptor to set
	 */
	public void setDescriptor(String descriptor) {
		this.descriptor = descriptor;
	}








	/**
	 * An informational string to appear in a status line when this snapshot is being displayed 
	 * 
	 * @return the descriptor
	 */
	public String getDescriptor() {
		return descriptor;
	}


	
	






	/**
	 * The corresponding source code location to be shown while this snapshot is being displayed.
	 * @param sourceLoc the sourceLoc to set
	 */
	public void setSourceLoc(SourceLocation sourceLoc) {
		this.sourceLoc = sourceLoc;
	}








	/**
	 * The corresponding source code location to be shown while this snapshot is being displayed.
	 * @return the sourceLoc
	 */
	public SourceLocation getSourceLoc() {
		return sourceLoc;
	}

	
	private class EntityIterator implements Iterator<Entity> {
		private Iterator<Identifier> mapIterator;
		private Iterator<Entity> listIterator;
		
		EntityIterator()
		{
			mapIterator = getEntities().keySet().iterator();
			listIterator = null;
		}

		@Override
		public boolean hasNext() {
			return mapIterator.hasNext() 
			  || (listIterator != null && listIterator.hasNext());
		}

		@Override
		public Entity next() {
			while (listIterator == null || !listIterator.hasNext()) {
				if (mapIterator.hasNext()) {
					Identifier id = mapIterator.next();
					listIterator = getEntities().get(id).iterator();
				} else {
					throw new NoSuchElementException();
				}
			}
			return listIterator.next();
		}

		@Override
		public void remove() {
			throw new UnsupportedOperationException();
		}
		
		
	}

	@Override
	public Iterator<Entity> iterator() {
		return new EntityIterator();
	}


	/**
	 * @param stackEntity the stackEntity to set
	 */
	public void setStackEntity(EntityIdentifier stackEntity) {
		this.stackEntity = stackEntity;
	}


	/**
	 * @return the stackEntity
	 */
	public EntityIdentifier getStackEntity() {
		return stackEntity;
	}



	/**
	 * @return the entities
	 */
	HashMap<Identifier, LinkedList<Entity> > getEntities() {
		return entities;
	}

	/**
	 * List the entities that have been registered as global variables
	 * @return
	 */
	public List<EntityIdentifier> getGlobals() {
		LinkedList<EntityIdentifier> globals = new LinkedList<EntityIdentifier>();
		for (Component comp: astack.getGlobalComponents()) {
			Object cobj = comp.getComponentObject();
			String clabel = comp.getLabel();
			if (cobj != null) {
				EntityIdentifier eid = new EntityIdentifier(Identifier.getIdentifierFor(cobj, astack.context()), clabel);
				globals.add(eid);
			}
		}
		return globals;
	}

	
}
