//
//  algae_go
//    This program is used to launch an AlgAE server, either as a CGI
//    call from a Web page or as a standalone application.  It is invoked
//    this way:
//      algae_go.cgi textfile server -name1 value1 -name2 value2 ... valueN
//    from the command line or as
//      http:// ... /algae_go.cgi?TEXT=textfile&SERVER=server&name1=value1...
//    in a web page.   In both cases, the (nameK,valueK) pairs are command
//    arguments for the AlgAE C++ server program.
//
//    algae_go does the following:
//     1) It obtains, from the machine it is running on, an unused port 
//        number for a TCP/IP socket.  For the purpose of discussion, call
//        this number P.
//     2) It then reads each line of the "textfile", replacing any occurrences
//        of the string "SOCKETNUM" by the port number P, and writing the
//        resulting lines to the standard output.  (If algae_go was invoked
//        via CGI, this textfile should be an HTML page containing the 
//        <APPLET> tag to launch the AlgAE client applet.  If algae_go was
//        invoked via the command line, the textfile could be a parameter
//        extension file for the AlgAE client standalone program.)
//     3) It spawns a child process that runs the program
//            server -port P -name1 value1 -name2 value2 ... valueN
//   
#include <algae/config.h>

#ifdef USE_WINSOCK
#include <winsock2.h>
#else
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netdb.h>
#endif

#include <errno.h>

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>

using namespace std;

void parseCGIArgs (const char* queryString, int& nArgs, char**& args);

ostream* debug;

#define NDEBUG
#define DEBUG if (debug != 0) (*debug)


int main (int nCommandArgs, char** commandArgs)
{
  //
  // With no command parameters, assume this was invoked via CGI from
  // a web page.
  bool invokedAsCGI = nCommandArgs <= 1;

  int nargs;
  char** args;
  
  if (invokedAsCGI)
    {
#ifdef NDEBUG
      debug = 0;
#else
      debug = new ofstream ("algae_go.log", ios::out);
#endif

      // Get CGI parameters
      char* cgiArgs;
      if ((cgiArgs = getenv("QUERY_STRING")) != 0)
	{
	  parseCGIArgs (cgiArgs, nargs, args);

	  DEBUG << "extracted CGI args: ";
	  for (int j = 0; j < nargs; ++j)
	    DEBUG << " (" << args[j] << ")";
	  DEBUG << endl;
	}
    }
  else
    {
#ifdef NDEBUG
      debug = 0;
#else
      debug = &cerr;
#endif

      // On the command line, args[0] is just the name of this program.
      // Ignore it.
      args = commandArgs + 1;
      nargs = nCommandArgs - 1;
    }
  

  // get socket info
#ifdef USE_WINSOCK
  WSADATA wsadata;
  if (WSAStartup(0x0101, &wsadata) == SOCKET_ERROR)
    {
      cerr << "Error: could not initialize WinSock 1.1" << endl;
      exit (-2);
    }
#endif
  struct  sockaddr_in server;


  // Construct the socket name
  memset (&server, 0, sizeof(server));
  server.sin_family = AF_INET;
  server.sin_addr.s_addr = htonl(INADDR_ANY);
  server.sin_port = htons(0);

  // Create the socket
  int sd = socket (AF_INET,SOCK_STREAM,0);
  
  if (sd<0) {
    cerr <<"Error opening stream socket"<<endl;
    exit(-1);
  }
  DEBUG <<"Socket created" << endl;


  // And bind it to this host
  if ( bind( sd, (struct sockaddr*) &server, sizeof(server)) < 0) {
    close(sd);
    cerr <<"Error: " << errno
	 << " binding name to stream socket"<<endl;
    exit(-1);
  }
 DEBUG <<"Socket bound" << endl;


  socklen_t length = sizeof(server);
  if ( getsockname (sd,(struct sockaddr*) &server, &length) ) {
    cerr <<"Error: getting socket name"<<endl;
    exit(-1);
  }
      


  int socketNum = ntohs(server.sin_port);
  DEBUG << "Socket number is " << socketNum << endl;


  // Get the port number as a string
  ostringstream socketNumOut;
  socketNumOut << socketNum;
  string socketNumString = socketNumOut.str();

  close (sd);

  //  Send the CGI output (usually an HTML page with an AlgAE
  //    client applet).

  const string socketnum = "SOCKETNUM";
  const int socketnumLen = socketnum.length();

  ifstream htmlstream (args[0]);
  string line;
		  
  while (getline (htmlstream, line))
    {
      // Replace any occurrences of socketnum by the actual port number.
      string::size_type pos = line.find (socketnum);
      while (pos >= 0 && pos < line.length())
	{
	  line.replace (pos, socketnumLen, socketNumString);
	  pos = line.find (socketnum);
	}
      
      cout << line << endl;
    }
  htmlstream.close();

  DEBUG <<"CGI output complete" << endl;


  // Build a new argument list that includes the port number.
  int newNargs = nargs + 2;
  char** newArgs = new char* [newNargs];
  for (int i = 2; i < nargs; ++i)
    newArgs[i+1] = args[i];
  newArgs[1] = "-port";
  newArgs[2] = strdup(socketNumString.c_str());;
  newArgs[nargs+1] = 0;

  char* programName = args[1];
  newArgs[0] = programName;
  
  DEBUG << "execv " <<  programName;
  for (int j = 0; j < newNargs-1; ++j)
    DEBUG << " (" << newArgs[j] << ")";
  DEBUG << endl;


  // Finally, spawn a child process to run the server
#ifdef USE_FORK
  if (debug != &cerr)
    delete debug;
  
  if (fork() == 0)
    {
      // This is the child process.
      int execStatus = execv (programName, newArgs);
      if (execStatus < 0)
	{
	  cerr <<"exec returned with status "
		<< execStatus << endl;
	}
    }
#else
  int execStatus = _spawnv (_P_NOWAIT, programName, newArgs);
  if (execStatus < 0)
    {
      cerr <<"exec returned with status "
	    << execStatus << endl;
    }
  DEBUG << "_spawnv status " << execStatus << endl;
  if (debug != &cerr)
    delete debug;
#endif

#ifdef USE_WINSOCK
  WSACleanup();
#endif

  return 0;
}





int getHexDigit(char c)
{
  if ((c >= '0') && (c <= '9'))
    return c - '0';
  else if ((c >= 'A') && (c <= 'F'))
    return c - 'A';
  else if ((c >= 'a') && (c <= 'f'))
    return c - 'a';
  else
    return 0;
}



/* Extract from the character array from[startPos..endPos-1] a string
   after applying the CGI conversion rules:
     All '+' characters are replaced by ' '
     All "%xx" sequences are replaced by the single character whose hex
       value is xx.
*/
string extractCGIString (const char* from, int startPos, int endPos)
{
  string out;

  for (int k = startPos; k < endPos; ++k)
    {
      if (from[k] == '+')
	out += ' ';
      else if (from[k] == '%')
	{
	  int digit1 = getHexDigit(from[k+1]);
	  int digit2 = getHexDigit(from[k+2]);
	  out += (char)(16*digit1 + digit2);
	  k += 2;
	}
      else
	out += from[k];
    }
  return out;  
}



/*  Parse a set of arguments passed via the CGI GET method, contained
    in the string queryString.  Return an array of argument strings
    (args) and a number of items in that array (nArgs) just as if the
    arguments had actually been passed via the command line.

    In particular, each CGI argument "...name=value..." is entered
    into the output array args as a pair of strings: "-name" and "value".
*/

void parseCGIArgs (const char* queryString, int& nArgs, char**& args)
{
  int queryLen = strlen(queryString);
  
  // First, how many strings will we need in the output array?

  int nPairs = 0;
  for (int i = 0; i < queryLen; ++i)
    if (queryString[i] == '=')
      nPairs += 1;

  nArgs = 2*nPairs;
  args = new char*[nArgs];

  // Now scan the query string, extracting each argument pair
  int scanPos = 0;
  string hyphen = "-";
  string name;

  int newPosition = 2;
  for (int pairNumber = 0; pairNumber < nPairs; ++pairNumber)
    {
      // Find the position of the '=' sign
      int equalsPos = scanPos;
      while ((equalsPos < queryLen) && (queryString[equalsPos] != '='))
	++equalsPos;

      
      // Find the position of the '&' terminator
      int ampersandPos = equalsPos;
      while ((ampersandPos < queryLen) && (queryString[ampersandPos] != '&'))
	++ampersandPos;

      // Extract the strings and place in the argument array
      string name0 = extractCGIString (queryString, scanPos, equalsPos);
      name = hyphen + name0;
      string value = extractCGIString (queryString, equalsPos+1, ampersandPos);
      char* valueC = strdup(value.c_str());

      if ((name0 == "TEXT") || (name0 == "text"))
	args[0] = valueC;
      else if ((name0 == "SERVER") || (name0 == "server"))
	args[1] = valueC;
      else 
	{
	  args[newPosition] = strdup (name.c_str());
	  args[newPosition+1] = valueC;
	  newPosition += 2;
	}

      scanPos = ampersandPos+1;
    }
}
