#include <algae/config.h>
#include IOSTREAM_INCL
#include FSTREAM_INCL
#include CPPSTRING_INCL

bool getLine(istream& in, string& line)
{
  const int MAXLENGTH = 1024;
  line = string();
  char ch;

  while (1)
    {
      if (in.get(ch))
	{
	  if (ch == '\n')
	    return true;
	  else if (ch == '\r')
	    {/* ignore return chars */}
	  else 
	    line += ch;
	  
	  if (line.length() > MAXLENGTH)
	    return false;
	}
      else
	return line.length() > 0;
    }
}

inline bool within(int x, int l, int h)
{
  return (x >= l) && (x < h);
}


int main (int nargs, char** args)
{
  if (nargs != 4) {
    cerr
<< "  filtersrc\n"
<< "\n"
<< "  This program can be used to prepare \"filtered\" source code\n"
<< "  and index files for display by AlgAE.\n"
<< "\n"
<< "  It is invoked from the command line as:\n"
<< "     filtersrc foo bar baz\n"
<< "  where \"foo\" is the name of an existing source code file (probably\n"
<< "  Java or C++, but the program doesn't really care).  \"bar\" and \"baz\"\n"
<< "  are the names for the two output files.  If files already exist with\n"
<< "  these names, they will be overwritten by this program.\n"
<< "\n"
<< "  The outputs are:\n"
<< "    \"bar\": a filtered source file.  This is identical to the \"foo\"\n"
<< "           file except that:\n"
<< "           1) In any line with the character sequence /*{*/, all\n"
<< "              characters beginning with that sequence up to the end of\n"
<< "              the line or a /*}*/ sequence are omitted.  For example,\n"
<< "              if the input file contains:\n"
<< "                 p = new Node(x); /*{*/ p->highlight(); /*}*/ q = p;\n"
<< "              then the output line would be:\n"
<< "                 p = new Node(x);  q = p;\n"
<< "           2) If a line begins with /*{*/ and, after applying rule 1, is\n"
<< "              completely empty (not even a blank character), then, no \n"
<< "              output line is generated at all.\n"
<< "           3) Tab characters are converted to spaces. Enough spaces are\n"
<< "              added to make the number of characters in the line evenly\n"
<< "              divisible by TABWIDTH.\n"
<< "\n"
<< "    \"baz\": the frame index file.  Any input line containing the string\n"
<< "           the string \"algae.FRAME\" or \"algae->FRAME\" is indexed.\n"
<< "           This string is assumed to be followed by a parenthesized\n"
<< "           parameter list.  The first parameter in this list is the\n"
<< "           frame identifier, and must be a string literal enclosed\n"
<< "           in \"\", with no \" inside.\n"
<< endl;
    exit(1);
    }


  const string FRAME_INDICATOR1 = "algae.FRAME";
  const string FRAME_INDICATOR2 = "algae->FRAME";
  const string BEGIN_OMISSION = "/*{*/";
  const string END_OMISSION = "/*}*/";
  const string STRING_LITERAL_DELIMITER = "\"";
  const int TABWIDTH = 8;



  ifstream insource (args[1]);
  if (!insource)
    {
      cerr << "Cannot open input file " << args[1] << endl;
      exit (1);
    }
  
  ofstream outsource (args[2]);
  if (!outsource)
    {
      cerr << "Cannot open output file " << args[2] << endl;
      exit (1);
    }

  ofstream outindex (args[3]);
  if (!outindex)
    {
      cerr << "Cannot open output file " << args[3] << endl;
      exit (1);
    }

  string line;
  int lineCounter = 0;
  
  while (getLine(insource, line))
    {
      int lineLen = line.size();
      bool wasOriginallyEmpty = (lineLen == 0);
      
      // First, look for frame commands.
      int pos = -1;
      int pos1, pos2;
      
      while (within (pos1 = line.find(FRAME_INDICATOR1, pos+1), 0, lineLen)
	     || within (pos2 = line.find(FRAME_INDICATOR2, pos+1), 0, lineLen))
	{
	  if (!within(pos1, 0, lineLen))
	    pos1 = lineLen + 1;
	  if (!within(pos2, 0, lineLen))
	    pos2 = lineLen + 1;
	  if (pos1 < pos2)
	    pos = pos1;
	  else
	    pos = pos2;
	  
	  int startString = line.find(STRING_LITERAL_DELIMITER, pos);
	  if (!within(startString, 0, lineLen))
	    {
	      cerr << "Frame in line " << lineCounter
		   << " not identified by a string literal." << endl;
	      continue;
	    }
	  
      	  int endString = line.find(STRING_LITERAL_DELIMITER, startString+1);
	  if (!within(endString, 0, lineLen))
	    {
	      cerr << "String literal for frame in line " << lineCounter
		   << " not terminated by \"." << endl;
	      continue;
	    }

	  string frameID = line.substr(startString+1, endString-startString-1);
	  outindex << lineCounter << " " << frameID << endl;
	}
      
      // Now, look for portions of line to be omitted.
      pos = 0;
      while (within(pos = line.find(BEGIN_OMISSION, pos), 0, lineLen))
	{
	  int n0 = pos+BEGIN_OMISSION.size();
	  int endPos = line.find(END_OMISSION, n0);
	  if (within(endPos, 0, lineLen))
	    {
	      n0 = endPos-pos+END_OMISSION.size();
	      line.STRERASE (pos, n0);
	    }
	  else
	    line.STRERASE (pos, lineLen);
	  lineLen = line.size();
	}
      
      if ((lineLen > 0) || wasOriginallyEmpty)
	{
	  // scan and replace tab characters
	  pos = 0;
	  const string tab = "\t";
	  const string blanks (TABWIDTH, ' ');
	  while (within(pos = line.find(tab, pos), 0, lineLen))
	    {
	      line.replace (pos, 1, blanks, 0, TABWIDTH-(pos % TABWIDTH));
	      lineLen = line.size();
	    }
	  outsource << line << endl;
	  ++lineCounter;
	}
    }
  return 0;
}
