#include <algae/messages.h>
#include <stdlib.h>

using namespace std;

const string BALGORITHM = "BALGORITHM";
const string BFRAME = "BFRAME";
const string COMPONENT = "COMP";
const string CONTINUING = "CONTINUING";
const string EDGE = "EDGE";
const string EALGORITHM = "EALGORITHM";
const string EFRAME = "EFRAME";
const string ITEMSELECTED = "ITEMSELECTED";
const string MENUITEM = "MENUITEM";
const string MESSAGE = "MESSAGE";
const string NODE = "NODE";
const string NOPAUSING = "NOPAUSING";
const string PAUSING = "PAUSING";
const string PROMPT = "PROMPT";
const string QUIT = "QUIT";
const string RESPONSE = "RESPONSE";
const string SERVERQUIT = "SERVERQUIT";
const string STATUS = "STATUS";
const string STDIN = "STDIN";
const string STDOUT = "STDOUT";
const string TITLE = "TITLE";


Messages::Messages (istream& inputs, ostream& outputs)
  : in(inputs), out(outputs), touchingEdges(true), pausing(true)  {}


bool Messages::getLine(string& line)
{
  const int MAXLENGTH = 1024;
  line = string();
  char ch;

  while (1)
    {
      if (in.get(ch))
	{
	  if (ch >= ' ')
	    line += ch;
	  else if (line.length() > 0)
	    return true;
	  if (line.length() > MAXLENGTH)
	    return false;
	}
      else
	return false;
    }
}


  // Inputs from the client
  //   [These may block until the client actually sends a signal. They may
  //    also terminate the program if the client shuts down unexpectedly.]


string Messages::decode(const string& str)
{
  string decoded;
  if ((str[0] == '"') && (str[str.size()-1] == '"'))
    decoded = str.substr(1, str.size()-2);
  else
    decoded = str;
  
  for (int i = 0; i < decoded.size(); ++i)
    {
      char ch = decoded[i];
      if (ch == '\\')
	{
	  int k;
	  char c = decoded[i+1];
	  if (c < 'A')
	    k = c - '0';
	  else
	    k = c - 'A' + 10;

	  c = decoded[i+2];
	  if (c < 'A')
	    k = 16*k + c - '0';
	  else
	    k = 16*k + c - 'A' + 10;

	  decoded.erase (i+1, 2);
	  decoded[i] = (char)k;
	}
    }
  return decoded;
}





bool Messages::canContinue()
{
  if (continuingPermitted)
    {
      continuingPermitted = false;
      return true;
    }
  else
    {
      string msg;
      while (getLine (msg))
	{
	  if (msg == CONTINUING)
	    {
	      return true;
	    }
	  else if (msg == PAUSING)
	    {
	      pausing = true;
	    }
	  else if (msg == NOPAUSING)
	    {
	      pausing = false;
	    }
	}
      return false;
    }
}



bool Messages::clientIsPausing() const
{
  return pausing;
}


	    
string Messages::getMenuSelection()
{
  string msg;
  while (getLine (msg))
    {
      if (msg == CONTINUING)
	{
	  continuingPermitted = true;
	}
      else if (msg.find(ITEMSELECTED) == 0)
	{
	  int pos = msg.find_first_not_of (" ", ITEMSELECTED.length());
	  continuingPermitted = true;
	  return msg.substr(pos);
	}
      else if (msg == QUIT)
	return QUIT;
      else if (msg == PAUSING)
	{
	  pausing = true;
	}
      else if (msg == NOPAUSING)
	{
	  pausing = false;
	}
    }
  return QUIT;
}

  
string Messages::promptForInput (const string& prompt)
{
  out << PROMPT << " " << encode(prompt) <<endl;
  
  string msg;
  while (getLine (msg))
    {
      if (msg == CONTINUING)
	{
	  continuingPermitted = true;
	}
      else if (msg.find(RESPONSE) == 0)
	{
	  int pos = msg.find ('"', RESPONSE.length());
	  int endPos = msg.rfind ('"');
	  string response = decode(msg.substr(pos+1, endPos-pos-1));
	  cerr << "Received response: /" << response << "/" << endl;
	  
	  return response;
	}
      else if (msg == QUIT)
	exit(1);
      else if (msg == PAUSING)
	{
	  pausing = true;
	}
      else if (msg == NOPAUSING)
	{
	  pausing = false;
	}
    }
  exit(1);
  return msg;
}


string Messages::stdInput ()
{
  out << STDIN << endl;
  
  string msg;
  while (getLine (msg))
    {
      if (msg == CONTINUING)
	{
	  continuingPermitted = true;
	}
      else if (msg.find(RESPONSE) == 0)
	{
	  int pos = msg.find ('"', RESPONSE.length());
	  int endPos = msg.rfind ('"');
	  return decode(msg.substr(pos+1, endPos-pos-1));
	}
      else if (msg == QUIT)
	exit(1);
      else if (msg == PAUSING)
	{
	  pausing = true;
	}
      else if (msg == NOPAUSING)
	{
	  pausing = false;
	}
    }
  exit(1);
  return msg;
}



  
//////////////////////////////

// Outputs to the client

string Messages::encode(const string& str)
{
  string encoded (str);
  int i = 0;
  while (i < encoded.size())
    {
      char ch = encoded[i];
      if ((ch < ' ') || (ch > 'z') || (ch == '"') || (ch == '\\'))
	{
	  string num = "\\";
	  unsigned chnum = (unsigned)ch;
	  int k = chnum / 16;
	  num += (char)((k <= 9) ? ('0' + k) : ('A' -10 + k));
	  k = chnum % 16;
	  num += (char)((k <= 9) ? ('0' + k) : ('A' -10 + k));
	  encoded.replace (i, 1, num);
	  i += 3;
	}
      else
	++i;
    }
  return '"' + encoded + '"';
}


void Messages::beginEdgeList()
{
  touchingEdges = true;
}

void Messages::endEdgeList()
{}

void Messages::beginComponentList()
{
  touchingEdges = false;
}

void Messages::endComponentList()
{}

void Messages::touch (int sourceID, int destID,
		      AlgAE::Directions dir, AlgAE::Color color,
		      const string& edgeLabel)
{
  if (touchingEdges)
    out << EDGE
	<< " " << sourceID
	<< " " << destID
	<< " " << (int)dir
	<< " " << (int)color.red
	<< " " << (int)color.green
	<< " " << (int)color.blue
	<< " " << encode(edgeLabel)
	<< "\n";
  else
    out << COMPONENT
	<< " " << sourceID
	<< " " << destID
	<< "\n";
}

  
void Messages::node (int objectID, AlgAE::Color color,
		     const string& label, bool vertical)
{
  out << NODE
      << " " << objectID
      << " " << encode(label)
      << " " << (int)vertical
      << " " << (int)color.red
      << " " << (int)color.green
      << " " << (int)color.blue
      << "\n";
}

  
void Messages::beginAlgorithm()
{
  out << BALGORITHM << "\n";
}


void Messages::endAlgorithm()
{
  out << EALGORITHM << endl;
}


void Messages::beginFrame()
{
  out << BFRAME << "\n";
}


void Messages::endFrame(const string& frameIdentifier)
{
  out << EFRAME
      << " " << encode(frameIdentifier)
      << endl;
}


void Messages::menuItem (const string& itemName)
{
  out << MENUITEM
      << " " << encode(itemName)
      << "\n";
}

void Messages::status (const string& statusMsg)
{
  out << STATUS
      << " " << encode(statusMsg)
      << "\n";
}
  
void Messages::algorithmName (const string& algName, const string& aboutString)
{
  out << TITLE
      << " " << encode(algName)
      << " " << encode(aboutString)
      << "\n";
}
  

void Messages::quit()
{
  out << SERVERQUIT
      << endl;
}


void Messages::message (const string& msg)
{
  out << MESSAGE
      << " " << encode(msg)
      << endl;
}



void Messages::stdOutput (const string& msg)
{
  out << STDOUT
      << " " << encode(msg)
      << endl;
}

