#include <algae/messages.h>
#include <algae/visible.h>
#include <set>
#include <algorithm>

using namespace std;

const string Visible::emptyLabel;

unsigned Visible::lastID = 1000;  // reserve ID's 0-999 for possible future
                                  // special purposes.


typedef std::set<int, std::less<int> > HighlightedEdges;


struct Visible::VisibleInfo 
{
  // The normal (unhighlighted) color for this object
  AlgAE::Color defaultColor;

  // Are components of this object arranged vertically?
  bool isVertical;  
  
  // The current color for this object.  Same as defaultColor unless this
  // object is highlighted.
  AlgAE::Color currentColor;
  
  // Name for this object.  Null if no name assigned.
  std::string* name;
  
  // Highlighting info for outgoing edges, by direction
  bool highlightDir[17];
  
  // Used during traversals of objects actually being shown.
  bool mark;
  
  HighlightedEdges highlightedEdges;
  
  VisibleInfo()
    : name(0), mark(false)
    {
      std::fill (highlightDir, highlightDir+17, false);
    }
};





Visible::Visible (AlgAE::Color color, bool vertical)
{
  vis = new VisibleInfo();
  vis->defaultColor = color;
  vis->currentColor = color;
  vis->isVertical = vertical;
  
  // Assign this object a unique ID
  ++lastID;
  ID = lastID;
  
  // Notify the client that this object exists.
  algae->addVisibleObject(this);
}

Visible::Visible (const Visible& v)
{
  vis = new VisibleInfo();
  vis->defaultColor = v.vis->defaultColor;
  vis->currentColor = v.vis->defaultColor;
  vis->isVertical = v.vis->isVertical;
  
  // Assign this object a unique ID
  ++lastID;
  ID = lastID;
  
  // Notify the client that this object exists.
  algae->addVisibleObject(this);
}

Visible::~Visible()
{
 algae->hideObject(this);
 algae->removeVisibleObject(this);
 delete vis;
}


AlgAE::Color Visible::edgeColor
    (const Visible* dest, AlgAE::Directions dir)
    const
{
    if (vis->highlightDir[dir])
      return AlgAE::Red;
    else
      {
        HighlightedEdges::iterator e = vis->highlightedEdges.find (dest->ID);
	if (e != vis->highlightedEdges.end())
	  return AlgAE::Red;
	else
	  return AlgAE::Black;
      }
}


void Visible::touch (const Visible* ptr, AlgAE::Directions dir,
                     const string& edgeLabel) const
{
  int destinationID = (ptr == 0) ? 0 : ptr->ID;
  algae->msgs->touch (ID, destinationID, dir, edgeColor(ptr, dir), edgeLabel);
  algae->enqueue (ptr);
}

void Visible::touch (const Visible& ref, AlgAE::Directions dir,
                     const string& edgeLabel) const
{
  int destinationID = ref.ID;
  algae->msgs->touch (ID, destinationID, dir, edgeColor(&ref, dir), edgeLabel);
  algae->enqueue (&ref);
}


void Visible::highlight() const
{
    vis->currentColor.red = 255 - vis->defaultColor.red;
    vis->currentColor.blue = 255 - vis->defaultColor.blue;
    vis->currentColor.green = 255 - vis->defaultColor.green;
}

void Visible::highlight(AlgAE::Color color) const
{
    vis->currentColor = color;
}

void Visible::unHighlight() const
{
    vis->currentColor = vis->defaultColor;
}

void Visible::clearHighlighting()
{
    vis->currentColor = vis->defaultColor;
    std::fill (vis->highlightDir, vis->highlightDir+17, false);
    vis->highlightedEdges.erase (vis->highlightedEdges.begin(),
				 vis->highlightedEdges.end());
}


void Visible::unHighlightAll()
{
    algae->unHighlightAll();
}


void Visible::highlight(AlgAE::Directions dir) const
{
    vis->highlightDir[dir] = true;
}

void Visible::unHighlight(AlgAE::Directions dir) const
{
    vis->highlightDir[dir] = false;
}


void Visible::highlight(const Visible* ptr) const
{
    if (ptr != 0)
      vis->highlightedEdges.insert (ptr->ID);
}

void Visible::unHighlight(const Visible* ptr) const
{
    if (ptr != 0)
      vis->highlightedEdges.erase (ptr->ID);
}

void Visible::setColor (AlgAE::Color color) const
{
    vis->currentColor = vis->defaultColor = color;
}

AlgAE::Color Visible::color () const
{
  return vis->currentColor;
}


void Visible::show() const
{
  algae->showObject (this);
}

void Visible::hide() const
{
  algae->hideObject (this);
}

void Visible::hideAllSince() const
{
  algae->hideObjectsSince (this);
}

void Visible::setName (const string& name) const
{
    vis->name = new std::string (name);
}

void Visible::setName (const char* name) const
{
    vis->name = new std::string (name);
}

std::string Visible::getName () const
{
  return (vis->name != 0) ? *(vis->name) : string("");
}


int Visible::id () const
{
  return ID;
}

bool Visible::vertical () const
{
  return vis->isVertical;
}


void Visible::setVertical (bool vert) const
{
  vis->isVertical = vert;
}


bool Visible::marked() const
{
  return vis->mark;
}

void Visible::setMark (bool flag) const
{
  vis->mark = flag;
}


bool Visible::LessThanID::operator() (const Visible& x, const Visible& y)
{
  return x.id() < y.id();
}
  


bool Visible::EqualID::operator() (const Visible& x, const Visible& y)
{
  return x.id() == y.id();
}
  

