package edu.odu.cs.zeil.AlgAE;

import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.InputStream;
import java.lang.ref.WeakReference;
import java.util.HashMap;
import java.util.concurrent.ArrayBlockingQueue;

import javax.swing.JApplet;
import javax.swing.JFrame;
import javax.swing.JMenuBar;

import edu.odu.cs.zeil.AlgAE.Client.Client;
import edu.odu.cs.zeil.AlgAE.Client.Message;
import edu.odu.cs.zeil.AlgAE.Client.DataViewer.Animator;
import edu.odu.cs.zeil.AlgAE.Client.IOViewer.SimulatedPrintStream;
import edu.odu.cs.zeil.AlgAE.Server.MenuFunction;
import edu.odu.cs.zeil.AlgAE.Server.MenuGenerator;
import edu.odu.cs.zeil.AlgAE.Server.Server;
import edu.odu.cs.zeil.AlgAE.Snapshot.ActivationStack;

/**
 *  This is the base class for all AlgAE animations.
 *  It provides an applet that can be displayed inline within
 *  a web page or that can pop up a separate window. The separate
 *  window can also be launched from an ordinary application.
 *  
 *  For applets, this behavior is controlled via the applet 
 *  parameter "inline". If set to 1, the display is inline.
 *  If set to 0, the display is in a separate window. If
 *  unspecified, defaults to INLINE_DEFAULT.
 *
 *  @author Steven J Zeil
 **/
public abstract class Animation extends JApplet implements AnimationContext, MenuGenerator
{
	static private final int INLINE_DEFAULT = 0;
	private int inlineDisplay;
	private boolean isAnApplet = true;
	
	private ArrayBlockingQueue<Message> messages;
	private ActivationStack aStack;
	private Client client;
	private Server server;

	private String theTitle;
	private boolean startPausing;
	private boolean started;
	
	private static HashMap<Thread, WeakReference<Animation> > instances
	 = new HashMap<Thread, WeakReference<Animation> >();

	public Animation (String title, boolean pausing)
	{
		theTitle = title;
		startPausing = pausing;
		aStack = new ActivationStack(this);
		messages = new ArrayBlockingQueue<Message>(3);
		server = new Server(this);
		started = false;
	}


	/**
	 *  Supply a message to appear in the Help..About dialog.
	 *  Typically, this indicates the origin of the source code
	 *  being animated and the name of the person who prepared the
	 *  animation.
	 **/
	public abstract String about();


	  /**
	   *  Called from buildMenu to register an initial action to be
	   *  run at the start of the animation, before any selections
	   *  from the menu.
	   */
	  public void registerStartingAction (MenuFunction action)
	  {
		  server.registerMenuItem (Client.StartingAction, action);
	  }

	  /**
	   *  Called from buildMenu to add an item to the Algorithm menu.
	   */
	  public void register(String menuItem, MenuFunction action) {
		  server.registerMenuItem (menuItem, action);
		  client.registerMenuItem (menuItem);
	  }

	/**
	 * Applet init action - build the GUI and get ready to go
	 */
	public void init ()
	{
		isAnApplet = true;
		inlineDisplay = INLINE_DEFAULT;
		String inlineParam = getParameter("inline");
		if (inlineParam != null) {
			try {
				inlineDisplay= Integer.parseInt(inlineParam);
			} catch (NumberFormatException ex) {}
		}

		if (inlineDisplay > 0) {
			JMenuBar menubar = new JMenuBar();
			setJMenuBar(menubar);

			client = new Client (this, menubar, null, startPausing);
			getContentPane().add(client);
		} else {
			runAsMain();
		}
		isAnApplet = true;
	}

	/**
	 * Applet start action - begin running the animation
	 */
	public void start ()
	{
		if (inlineDisplay > 0) {
			if (!started) {
				started = true;
				client.start();
				server.start();
			} else {
				client.resume();
			}
		}
	}


	/**
	 * Applet stop action - pause the animation
	 */
	public void stop ()
	{
		if (inlineDisplay > 0) {
			client.pause();
		}
	}

	/**
	 * Applet destroy action - shut everything down for good
	 */
	public void destroy()
	{
		client.destroy();
	}

	
	/**
	 * Used to run this as a standalone application from main() rather than as an applet
	 */
	public void runAsMain()
	{
		inlineDisplay = 0;
		JFrame window = new JFrame(theTitle);
		JMenuBar menubar = new JMenuBar();
		window.setJMenuBar(menubar);

		client = new Client (this, menubar, window, startPausing);
		window.getContentPane().add(client);
		
		window.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				client.destroy();
				if (isAnApplet) {
					setVisible(false);
				} else {
					System.exit (0);
				}
			}
		});
		window.pack();
		isAnApplet = false;
		window.setVisible(true);
		//window.setPreferredSize(new Dimension(800, 600));
		server.start();
		client.start();
	}

	/**
	 * Each animation has a unique activation stack.
	 * @return the activation stack for this animation
	 */
	public ActivationStack getActivationStack() {
		return aStack;
	}
	
	/**
	 * Each animation has a unique client.
	 * @return the client for this animation
	 */
	public Client getClient ()
	{
		return client;
	}

	/**
	 * Each animation has a unique server.
	 * @return the server for this animation
	 */
	public Server getServer ()
	{
		return server;
	}

	/**
	 * Each animation has a queue of messages used to send
	 * info from the server to the client.
	 * 
	 * @return the message queue for this animation
	 */
	public ArrayBlockingQueue<Message> getMessageQueue ()
	{
		return messages;
	}

	/**
	 * Animated code must be able to access the relevant Animation instance
	 * even though such code was written independently of the animation system.
	 * This function returns that animation instance, under the assumption that the
	 * server (which launches the animated code) will have registered its thread
	 * for that purpose.
	 *  
	 * @return the animation associated with a thread
	 */
	private static Animation algae()
	{
		return instances.get(Thread.currentThread()).get();
	}
	
	/**
	 * Animated code must be able to access the relevant Animation instance
	 * even though such code was written independently of the animation system.
	 * This function associates this animation instance with a thread, usually
	 * the thread used by the server to launch the code being animated.
	 */
	public void registerInstance(Thread t)
	{
		instances.put(t, new WeakReference<Animation>(this));
	}

	
	/**
	 * This must be called at the beginning of each new function to signal
	 * that a new empty record should be pushed onto the top of the stack.
	 * 
	 * @param thisObj - A reference to an object of the class of
	 *    which the current function is a member. Normally, "this"
	 *    will do just fine.  thisObj is used to help locate
	 *    the source code being animated, so, in a pinch (e.g., if
	 *    animating a static function) another object whose source
	 *    code lies in the same directory/folder will do.
	 *    
	 * @return ActivationRecord for the new function call
	 */

	public static ActivationRecord activate (Object thisObject) {
		return algae().getActivationStack().activate (thisObject);
	}


	/**
	 * Show a variable as a global value in all displays.
	 * Variables portrayed by this call are shown "in-line".
	 * 
	 * @param label  the variable name (optional, can be "" or null)
	 * @param param  the variable/value
	 */
	public void globalVar(String label, int value)
	{
		getActivationStack().globalVar(label, value);
	}

	/**
	 * Show a variable as a global value in all displays.
	 * Variables portrayed by this call are shown "in-line".
	 * 
	 * @param label  the variable name (optional, can be "" or null)
	 * @param param  the variable/value
	 */
	public void globalVar (String label, Object param)
	{
		getActivationStack().globalVar(label, param);
	}

	
	
	/**
	 * Show a variable as a global value in all displays.
	 * Variables portrayed by this call are shown as labeled
	 * pointers to the actual value.
	 * 
	 * @param label  the variable name (optional, can be "" or null)
	 * @param param  the variable/value
	 * @return a reference to this breakpoint
	 */
	public void globalRefVar (String label, Object param)
	{
		getActivationStack().globalRefVar(label, param);
	}


	/**
	 * Pops up a dialog box prompting for an input, pausing the
	 * animation until a satisfactory input value is obtained from the
	 * human operator.
	 * 
	 * @param prompt  Text of the prompt message to be displayed
	 * @param requiredPattern regular expression describing an acceptable input value
	 * @return a human-entered string matching the requiredPattern
	 */
	public String promptForInput(String prompt, String requiredPattern) {
		return server.promptForInput(prompt, requiredPattern);
	}
	
	/**
	 * Pops up a dialog box prompting for an input, pausing the
	 * animation until a satisfactory input value is obtained from the
	 * human operator.
	 * 
	 * @param prompt  Text of the prompt message to be displayed
	 * @return a human-entered string
	 */
	public String promptForInput(String prompt) {
		return promptForInput(prompt, ".*");
	}
	
	
	/**
	 * Provides access to the Animator - useful mainly if an animation
	 * wishes to set the initial speed or pause/run mode
	 */
	protected Animator getAnimator() {
		return client.getAnimator();
	}
	
	/**
	 * A substitute for System.in - reads from the animation GUI
	 */
	public InputStream in;
	
	/**
	 * A substitute for System.out - writes to the animation GUI
	 */
	public SimulatedPrintStream out = new SimulatedPrintStream();
	
}



